/********************************************************************

    Plik:       eta.h

    Zawarto:  Deklaracje dla pliku etad.c


    Autor:      Marek Olejniczak


****************************************************************/


#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE !FALSE
#endif


/*  
 * Plik niezbdny dla SysVInit
 */

#define PWRSTAT "/etc/powerstatus"




/*  
 * Stany UPSa
 */

#define S_NOCHANGE -1  /* stan nieznany  */
#define S_OK        0  /* wszystko OK */
#define S_ONBAT     1  /* zasilanie bateryjne */
#define S_LOBAT     2  /* niski pozoim napicia baterii, trzeba wyczy! */
#define S_ERROR     4  /* bd UPSa - najprawdopodobniej niepodczony kabel */
#define S_VPPERROR  8  /* bd UPSa - nie mona ustawi +12V na zczu */
#define S_NOETAUPS  16 /* bd w nagwku pakietu sieciowego */
#define S_HEADERROR 32 /* za liczba skadnikw w nagwku sieciowym */


/*  
 *  Bity, ktre monitorujemy na porcie szeregowym
 */

#define B_VPLUS TIOCM_RTS         /* zasilanie = +12V -> jedynka logiczna */
#define B_KILLUPS  TIOCM_DTR      /* wyczenie UPSa */
#define B_ONBAT TIOCM_CTS         /* AC FAIL - brak zasilania sieciowego */
#define B_LOBAT TIOCM_CD          /* LBAT = Low Battery */
#define B_ERR   TIOCM_DSR         /* le podczony kabel lub jego brak */

/* Nagwek informacji o statusie przesyany w sieci */
char *NETHEADER = "ETAUPS";



/* Kody przekazywane do funkcji notify_init() */

#define  I_OK      0        /* UPS OK */
#define  I_ACFAIL  1        /* brak zasilania sieciowego */
#define  I_LOWBAT  2        /* niski poziom baterii */



/*-------------------------------------------------------------------*
     Informuje INIT o zaniku zasilania poprzez przesanie
     sygnau SIGPWR.

     Argumenty:   zgodne z typem wyliczeniowym Kodinit
 *-------------------------------------------------------------------*/

void notify_init(int kod);

/*-------------------------------------------------------------------*
  Rozwija status UPSa i zapisuje zmiany statusu do logu. Jeeli praca
  jest na baterii, informuje init aby rozpocz shutdown. Polecenie
  shutdown jest opniane o liczb Mode_Count cykli odczytu UPSa, aby
  zlikwidowa problem krtkich zanikw zasilania i nie zawraca gowy
  uytkownika takim stanem. Jeeli wykryty zostanie zanik zasilania
  sieciowego, to nastpuje skrcenie okresu odpytywania UPSa do
  wartoci Mode_Delay_Fail i jednoczenie nastpuje zmniejszenie
  staej Mode_Count do wartoci jeden.


     Argumenty:  Fail - kod statusu UPSa
 *-------------------------------------------------------------------*/

void EvalStatus (int Fail);




/*-------------------------------------------------------------------
     LogError

     Logowanie informacji o bdach do stderr lub syslog w zalenoci od
     tego, czy jestemy demonem czy te nie.
     

     Argumenty:  ModName - nazwa moduu zgaszajcego bd
                 Error  - informacja o bdzie

-------------------------------------------------------------------*/

void LogError (char *ModName, char *Error);


/*-------------------------------------------------------------------
     Usage

     Drukuje help i sposb wywoania programu

-------------------------------------------------------------------*/

void Usage (void);



/*-------------------------------------------------------------------
     NET_GetServerAddr

     Przeksztaca nazw serwera lub adres IP w formacie kropkowym
     na adres internetowy w sieciowej kolejnoci bajtw.

     Argumenty:  Server - nazwa serwera lub IP srewera

     Zwraca:     Adres serwera lub IN_ADDRNONE gdy bd
-------------------------------------------------------------------*/

unsigned long NET_GetServerAddr (char *Server);



/*-------------------------------------------------------------------
     NET_Check

     Pobiera status odlegego UPSa.

     Argumenty:  Server - adres serwera f formacie sieciowym
                  Port   - odlegy port pod ktry si podczamy

     Zwraca:     S_* kod statusu
-------------------------------------------------------------------*/

int NET_Check (unsigned long Server, int Port);


/*-------------------------------------------------------------------
     NET_Serve

     Udostepnia status UPSa do odlegego hosta w sieci

     
     Argumenty:  Port    - numer portu na ktrym nasuchujemy
                  MaxWait - maksymalny czas osczekiwania na poczenie 
		            w sekundach
     
     Zwraca:
-------------------------------------------------------------------*/

void NET_Serve (int Port, int MaxWait);




/*-------------------------------------------------------------------
     UPS_Open

     Otwiera poczenie z UPSem przez port szergowy.

     Argumenty:  DevName - nazwa portu szeregowego do ktrego podczony 
                           jest UPS

     Zwraca:     deskryptor otwartego poczenia
-------------------------------------------------------------------*/

int UPS_Open (char *DevName);


/*-------------------------------------------------------------------
     ups_setup

     
     Inicjuje port dla celw monitorowania UPSa (ustawia odpowiednie 
     linie, itp). Umoliwia wyaczenie UPSa.

     Argumenty:  fd   - deskryptor portu (zwrcony przez UPS_Open)
                 kill - TRUE aby wyczy UPS

-------------------------------------------------------------------*/

void ups_setup (int fd, int kill);


/*-------------------------------------------------------------------
     ups_check

     Sprawdza ststus UPSa (OK, On Battery lub Low Battery)
     

     Argumenty:  fd   - deskryptor portu (zwrcony przez UPS_Open)

     Zwraca:     S_* kod statusu
-------------------------------------------------------------------*/

int ups_check (int fd);






