/********************************************************************

    Monitorowanie UPSa firmy ETA w celu wykrycia stanu zaniku
    zasilania sieciowego lub niskiego poziomu baterii.
    Moe pracowa jako serwer w celu nadzorowania UPSa podczonego 
    do komputera lub moe monitorowa poprzez sie stan odlegego UPSa 
    (klient).
    W przypadku wykrycia niskiego poziomu baterii natychmiast wycza 
    komputer.
                

Autor:  Marek Olejniczak ( marekol@xox.pl )


**********************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "eta.h"




#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif


/*  
 *  Flagi ustawiane przez opcje linii polece
 */

static int Mode_Cable = TRUE;  /* czy testowa podczenie kabla monitorujcego */
static int Mode_Kill;
static int Mode_NoMaster;
static int Mode_LowBat;
static int Mode_Test;
static int Mode_Port = 401;
static int Mode_Delay = 60;   /* okres odpytywania UPSa w stanie normalnej pracy */
static int Mode_Delay_Fail = 10;  /* okres odpytywania UPSa przy LOBAT */
static int Mode_Slave;
static int Mode_Count = 2;


static char *Version = "v1.0";  /* numer wersji programu */
static int ServerFD = -1;       /* deskryptor pliku dla trybu master */


char *ProgName = "etad";   /* nazwa programu dla syslog */
int   Status;     /* status UPSa */
int   IsDaemon;   /* 0 = nie jest demonem, 1 = demon, 2 = demon z logowaniem */



/*-------------------------------------------------------------------*


 Informuje INIT o zaniku zasilania poprzez przesanie sygnau SIGPWR.

 UWAGA! Aby init wykry kod LOW naley uy sysvinit-2.76 lub nowszego.

 *-------------------------------------------------------------------*/

void notify_init(int kod)
{
    int fd;


    /* 
     * Utworzenie pliku info wymaganego przez init
     */

    unlink (PWRSTAT);
    
    if ((fd = open (PWRSTAT, O_CREAT|O_WRONLY, 0644)) >= 0) {
      if (kod == I_OK) 
	  write (fd, "OK\n", 3);
      else if (kod == I_ACFAIL)
	  write (fd, "FAIL\n", 5);
      else if (kod == I_LOWBAT)
	  write (fd, "LOW\n", 4);
      else { /* to nigdy nie powinno si zdarzy ! */
	LogError("notify_init", "Powazny blad aplikacji, przerwanie programu");
	exit(1);
      }
        
      close (fd);
    }

    /*  
     * Wysanie sygnau, jeeli nie jestemy w trybie testowym
     */

    if (!Mode_Test)
        kill (1, SIGPWR);
}


/*-------------------------------------------------------------------*

  Rozwija status UPSa i zapisuje zmiany statusu do logu. Jeeli praca
  jest na baterii, informuje init aby rozpocz shutdown. Polecenie
  shutdown jest opniane o liczb Mode_Count cykli odczytu UPSa, aby
  zlikwidowa problem krtkich zanikw zasilania i nie zawraca gowy
  uytkownika takim stanem. Jeeli wykryty zostanie zanik zasilania
  sieciowego, to nastpuje skrcenie okresu odpytywania UPSa do
  wartoci Mode_Delay_Fail i jednoczenie nastpuje zmniejszenie
  staej Mode_Count do wartoci jeden.


     Argumenty:  Fail - kod statusu UPSa

 *-------------------------------------------------------------------*/

void UPS_Status(int Fail)
{
    static int Count;
    static int LastFail = S_NOCHANGE;
    static int oldcount;  /* przechowuje Mode_Count gdy ON BATTERY */
    static int oldmode_delay; /* przechowuje Mode_Delay gdy ON BATTERY */
    static int on_battery = 0; /* gdy TRUE to praca na baterii */


    ++Count;
    
    if (LastFail != Fail) {
        LastFail = Fail;
        Count = 0;
    }

    if ( (Count == Mode_Count) && (Fail != S_NOCHANGE) ) {
        
            Status = Fail; /* globalne rozgoszenie statusu UPSa */

	    /*
	     * Moliwe sytuacje:
	     *    S_VPPERROR - powany bd, nic nie da si zrobi!
	     *    S_ERROR - odczony kabel, nic nie da si zrobi!
	     *    S_OK - wszystko OK, nic do roboty
	     *    S_ONBAT - brak zasilania sieciowego
	     *    S_ONBAT & S_LOBAT - koczy si bateria i nie ma sieci,
	     *               stan krytyczny, trzeba wycza natychmiast
	     *    S_LOBAT - niskie napicie baterii ale jest sie, czekaj
	     */

	    if ( ! Fail ) { /* Jest wszystko OK! */
	      LogError("log", "UPS OK");
	      if ( on_battery ) {
		on_battery = 0;
		Mode_Count = oldcount;
		Mode_Delay = oldmode_delay;
		notify_init(I_OK); /* sie OK - odwoaj shutdown */
	      }
	    }
	    else
	    if ( Fail & S_VPPERROR ){
	      /*
	       * Powany bd UPSa (lub raczej cza szeregowego w kompie!)
	       */
	      LogError ("log", "UPS Panika - brak +12V");
	    }
	    else
	    if ( Fail & S_ERROR ){ /* Niepodczony kabel */
	      LogError ("log", "UPS Panika - kabel odlaczony");
	    }
	    else
	    if ( Fail & S_LOBAT ){ /* Niskie napicie baterii !!! */
	      if ( Fail & S_ONBAT ) {/* Praca na baterii */
		LogError ("log", "Niskie napiecie baterii");
		notify_init(I_LOWBAT);
	      }
	      else {
		LogError ("log", "Powrot zasilania i niskie napiecie baterii");
		notify_init(I_OK); /* Jest sie mimo i jest niska bateria */
	      }
	    }
	    else
	    if ( Fail & S_ONBAT ){ /* Praca na baterii */
	      on_battery = 1;  /* zaznacz, e jeste na baterii */
	      oldcount = Mode_Count;
	      oldmode_delay = Mode_Delay;
	      Mode_Count = 1; /* nowa warto licznika */
	      Mode_Delay = Mode_Delay_Fail; /* nowy okres prbkowania */
	      LogError ("log", "Praca na baterii");
	      if (!Mode_LowBat)
		notify_init(I_ACFAIL);
	    }
	    else
	    if ( (Fail & S_NOETAUPS) && Mode_Slave )
	      /* Bd w nagwku odpowiedzi serwera, chyba zy serwer */
	      LogError ("log", "Blad - odleglym serwerem nie jest etad");
	    else
	    if ( (Fail & S_HEADERROR) && Mode_Slave )
	      /* Bd w odpowiedzi zdalnego serwera */
	      LogError ("log", "blad w odpowiedzi zdalnego serwera");
	    else {/* to nigdy nie powinno si zdarzy */
	      LogError ("log", "UPS Panika - nieprawidlowy status");
	      if ( Mode_Test )
		fprintf(stderr, "Kod paniki = %d\n", Fail);
	    }	      
    } /*if()*/
}



/*-------------------------------------------------------------------*
     LogError

     Logowanie wiadomoci do stderr lub syslog w zalenoci od tego, 
     czy program zosta uruchomiony jako demon.

     Argumenty:  Module - nazwa moduu zgaszajcego bd
                 Error  - info o bdzie

 *-------------------------------------------------------------------*/

void LogError (char *Module, char *Error)
{
    if (IsDaemon) {
        if (IsDaemon == 1) {
            openlog (ProgName, LOG_CONS, LOG_DAEMON);
            ++IsDaemon;
        }
        syslog (LOG_WARNING, "%s: %s", Module, Error);
    }
    else
      fprintf (stderr, "%s: %s: %s\n", ProgName, Module, Error);

}


/*-------------------------------------------------------------------*
     Parametry

     Drukowanie helpu i info o programie

 *-------------------------------------------------------------------*/

void Parametry (void)
{
    printf ("\n");
    printf ("---------------------------------------------------------------------\n");
    printf ("etad, wersja %s kwiecien 2000, Marek Olejniczak (marekol@xox.pl)\n", Version);
    printf ("Wywolanie:\n");
    printf ("    %s [opcje] port_szeregowy | host\n", ProgName);
    printf ("Opcje:\n");
    printf (" -h        Pomoc (ten tekst)\n");
    printf (" -c pauza  Opusc 'pauza' testow UPSa zanim wywolasz shutdown (domyslnie 2)\n");
    printf (" -k        Nie testuj polaczenia kabla monitorujacego\n");
    printf (" -i czas   Testuj stan UPSa co 'czas' sekund (domyslnie 60 sek.)\n");
    printf (" -w        Wylacz UPS\n");
    printf (" -l        Shutdown tylko wtedy, gdy niski poziom baterii\n");
    printf (" -m        Wylacz tryb master\n");
    printf (" -p port   Uzywaj 'port' dla celow komunikacji sieciowej (domyslnie 401)\n");
    printf (" -s        Startuj w trybie slave i monitoruj zdalny UPS\n");
    printf (" -t        Tryb testowy\n");
    printf ("---------------------------------------------------------------------\n");
}


/*-------------------------------------------------------------------*
     UPS_Open

     Otwarcie postu szeregowego do ktrego podaczony jest UPS

     Argumenty:  DevName - nazwa urzdzenia do otwarcia ( np. /dev/ttyS0 )

     Zwraca:     deskryptor otwartego pliku urzdzenia
 *-------------------------------------------------------------------*/

int UPS_Open (char *DevName)
{
    int fd;
    
    if ((fd = open (DevName, O_RDWR | O_NDELAY)) < 0)
        LogError ("opendev", sys_errlist [errno]);

    return fd;
}


/*-------------------------------------------------------------------*
     ups_setup

     Inicjacja portu dla celw monitorowania UPSa. Gdy kill=1 to
     nastpi wyczenie UPSa.
   

     Argumenty:  fd   - deskryptor dla urzdzenia szeregowego
                 kill - gdy rwny 1 to wyczenie zasilania UPSa

 *-------------------------------------------------------------------*/
void ups_setup (int fd, int kill)
{
    int set_bits;

    if ( !kill ) {
	set_bits = B_VPLUS;
	ioctl(fd, TIOCMSET, &set_bits); /* ustawiam zasilanie na czu */
        sleep (1);
    }
    else {
        set_bits = B_KILLUPS;
        LogError ("kill", "Wylaczam UPS!\n");
        ioctl(fd, TIOCMSET, &set_bits);
        sleep (10);
	/* nie doczekaem si wyczenia! */
        set_bits = B_VPLUS;
	ioctl(fd, TIOCMSET, &set_bits); /* ustawiam zasilanie na czu */
	sleep (1);
	
	/* Hmmm..... Nadal mam zasilanie, nie mog wyczy UPSa ! */

	LogError ("kill", "Nie moge wylaczyc UPSa!\n");

        exit (0);
    }
}


/*-------------------------------------------------------------------*
     ups_check


     Sprawdza status UPSa (OK, On Battery, Low Battery lub UPS 
     niepodaczony) poprzez testy linii DSR,DCD,CTS i RTS.
     
     Argumenty:  fd - deskryptor do portu UPSa

     Zwraca:     kod statusu
 *-------------------------------------------------------------------*/

int ups_check (int fd)
{
    int flags,status;
  
  
    ioctl (fd, TIOCMGET, &flags); /* czytam stan UPSa */
    if ( ~flags & B_VPLUS )
      ups_setup(fd, 0);  /* ponowna inicjacja cza UPSa */

    ioctl (fd, TIOCMGET, &flags); /* czytam ponownie stan UPSa */
    if ( ~flags & B_VPLUS )
      return S_VPPERROR;

    /*
     * Jest na zczu +12V, skadam status UPSa
     */

    status = S_OK;
    if ( (~flags & B_ERR) && Mode_Cable )
      status += S_ERROR;  /* jest odczony kabel */

    if ( ~flags & B_LOBAT )
      status += S_LOBAT;  /* jest niski poziom baterii */

    if ( ~flags & B_ONBAT )
      status += S_ONBAT;  /* brak zasilania sieciowego */

    return status; 
}



/*-------------------------------------------------------------------*
     SigTerm

     Obsuga sygnaw SIGTERM, SIGINT oraz SIGCHLD.

     Argumenty:  sig - numer odebranego sygnau

 *-------------------------------------------------------------------*/

void SigTerm (int Sig)
{
    if (Sig == SIGCHLD) {
        signal (SIGCHLD, SigTerm);
        while (waitpid (-1, NULL, WNOHANG) > 0)
            ;
    }
    else {
      if (ServerFD >= 0) {
        close (ServerFD);
        exit (0);
      }
    }
}


/*-------------------------------------------------------------------*
     TCP_Open

     Otwiera gniazdo klienta w celu skomunikowania si z serwerem.

     Argumenty:  HostAddr - adres odlegego hosta (serwera) w formacie 
                 sieciowym

                 Port     - numer zdalnego portu do ktrego si doczamy

     Zwraca:     deskryptor otwartego poczenia lub -1 gdy bd
 *-------------------------------------------------------------------*/

int TCP_Open (unsigned long HostAddr, int Port)
{
    int fd;
    struct sockaddr_in server_addr;
    
    bzero (&server_addr, sizeof (struct servent));
    server_addr.sin_family = AF_INET;

    if (Port <= 0) {
      LogError ("tcp_open", "Brak numeru portu");
      return -1;
    }

    server_addr.sin_port = htons (Port);

    if ((fd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
      LogError ("tcp_open", "Nie moge utworzyc gniazdka");
      return -1;
    }

    bcopy (&HostAddr, &server_addr.sin_addr, sizeof (HostAddr));
        
    if ( connect(fd, (struct sockaddr *) &server_addr, 
		 sizeof(server_addr)) < 0 ) {
      LogError ("tcp_open", "Brak polaczenia z serwerem");
      return -1;
    }

    return fd;
}


/*-------------------------------------------------------------------*
     ReadLine

     Czyta lini danych z odlegego hosta.

     Argumenty:  FD     - deskryptor gniazdka do odczytu
                 Buffer - bufor do ktrego zapisujemy odczytan lini
                 Len    - rozmiar bufora

     Zwraca:     Liczb odebranych bajtw
 *-------------------------------------------------------------------*/

int ReadLine (int FD, char *Buffer, int Len)
{
    int  i;
    int  rc;
    char c;

    for (i = 0; i < Len; ++i) {
        if ((rc = read (FD, &c, 1)) == 1) {
             *Buffer++ = c;
             if (c == '\n')
                 break;
        }
        else 
	if (rc == 0) {
            if (i == 1)
                return 0;
            else
                break;
        }
        else
          return -1;
    }
    
    *Buffer = 0;
    return i;
}


/*-------------------------------------------------------------------*
     NET_GetServerAddr

     Przetwarza nazw serwera lub jego adres IP (liczbowy) na adres
     internetowy w formacie sieciowym.

     Arguenty:  Server - nazwa serwera lub adres IP serwera

     Zwraca:    adres serwera (sieciowy) lub IN_ADDRNONE gdy bd
 *-------------------------------------------------------------------*/

unsigned long NET_GetServerAddr (char *Server)
{
    struct hostent *hp;
    unsigned long  inaddr;
    
    if ((inaddr = inet_addr (Server)) == INADDR_NONE) {
        if ((hp = gethostbyname (Server)) == NULL) {
            LogError ("get_addr", "blad w nazwie hosta");
            return -1;
        }

        bcopy (hp->h_addr, &inaddr, hp->h_length);
    }

    return inaddr;
}


/*-------------------------------------------------------------------*
     NET_Check

     Pobiera ststus UPSa z odlegego hosta

     Argumenty:  Server - adres serwera w formacie sieciowym
                 Port   - numer odlegego portu do podaczenia si

     Zwraca:     status odlegego UPSa
 *-------------------------------------------------------------------*/

int NET_Check (unsigned long Server, int Port)
{
    char   buffer [128];
    char   header [128];
    int status = S_NOCHANGE;
    int fd     = TCP_Open (Server, Port);

    if (fd >= 0) {
      /*
       * czytam status z odlegego hosta
       */

      if ( ReadLine(fd, buffer, sizeof (buffer)) > 0 ) {

	/* sscanf() przydziela dynamicznie bufor dla *header */
	if ( sscanf(buffer, "%s %d", header, &status ) == 2 ) {
	  if ( strcmp(header, NETHEADER) ) {
	    status = S_NOETAUPS; /* zy nagwek */
	    if ( Mode_Test )
	      fprintf(stderr, "Naglowek z serwera: %s\n", header );
	  }
	}
	else
	  status = S_HEADERROR; /* za liczba skadnikw w funkcji sscanf */

    } /* if ( ReadLine() ) */

      close (fd);
   } /* if (fd >=0 */

    return status;
}


/*-------------------------------------------------------------------*
     NET_Serve

     Dostarcza statusu lokalnego UPSa do odlegych komputerw w sieci.
     
     Argumenty:  Port    - numer portu do ktrego si podczamy
                 MaxWait - maksymalny czas oczekiwania na poczenie

 *-------------------------------------------------------------------*/

void NET_Serve (int Port, int MaxWait)
{
  static struct sockaddr_in serv_addr;
  static struct sockaddr_in cli_addr;
  fd_set fdset;
  int cli_len,newfd,sts,child;
  long endtime;
  struct timeval timeout;
  char *net_message;
    

    if (ServerFD < 0) {
        bzero (&serv_addr, sizeof (serv_addr));
        serv_addr.sin_family      = AF_INET;
        serv_addr.sin_addr.s_addr = htonl (INADDR_ANY);
        serv_addr.sin_port        = htons (Port);
        
        if ( (ServerFD = socket (AF_INET, SOCK_STREAM, 0)) < 0 ) {
            LogError ("server", "Nie moge otworzyc gniazda");
            exit (1);
        }

        sts = 1;
        
        setsockopt(ServerFD, SOL_SOCKET, SO_REUSEADDR, &sts, sizeof (sts));
        
        if ( bind (ServerFD,(struct sockaddr *) &serv_addr, 
		  sizeof (serv_addr)) < 0 ) {
            LogError ("server", "blad funkcji bind()");
            close (ServerFD);
            ServerFD = -1;
            exit (1);
        }

        /*  
	 * Definiowanie uywanych sygnaw i nasuchiwanie na otwartym gniazdku
         */

        listen (ServerFD, 5);
	signal (SIGTERM, SigTerm);
        signal (SIGINT,  SigTerm);
        signal (SIGCHLD, SigTerm);        
    } /* if (ServerFD < 0) */

    /*  
     * Oczekiwanie na podczenie si klienta. Nie czekaj na klienta duej
     * ni MaxWait sekund.
     */

    endtime = MaxWait + time (NULL);
    sts     = 1;

    while (sts > 0) {
        FD_ZERO (&fdset);
        FD_SET  (ServerFD, &fdset);

        timeout.tv_sec  = endtime - time (NULL);
        timeout.tv_usec = 0;

        if (timeout.tv_sec < 0)
            break;
        else
            sts = select (ServerFD + 1, &fdset, 0, 0, &timeout);

        if (sts < 0 && errno == EINTR) {
            /*  
	     * Przerwane wywoanie systemowe, najprawdopodobniej potomek si
	     * zakoczy wykonywa
             */

            sts = 1;
            continue;
        }
        else 
	if (sts > 0) {
            /*  
	     * Mamy klienta, wykonaj fork() i przelij status
             */

            cli_len = sizeof (cli_addr);
            newfd = accept (ServerFD, (struct sockaddr *) &cli_addr, &cli_len);
    
	    child = fork();
	    if ( child == 0 ) { /* jestem potomkiem, obsuga klienta */
	      asprintf(&net_message, "%s %d", NETHEADER, Status);
	      write( newfd, net_message, strlen(net_message) );
	      free( net_message );
              exit (0);  
	    }
	    else
            if (child > 0) /* jestem rodzicem */
                close (newfd);
            else {
	      LogError ("server", "blad funkcji fork()");
              return;
            }
        } /* if( sts > 0 ) */
    
	return;
    } /* while */
} 


/*-------------------------------------------------------------------*

 *********************  MAIN  PROGRAM  *****************************
  
 *-------------------------------------------------------------------*/

int main (int argc, char *argv [])
{
    int option;
    int fd     = -1;
    unsigned long inaddr = (unsigned) -1;

    /*  
     * Pobranie przecznikw z linii polece.
     */

    while ((option = getopt (argc, argv, "wktli:shp:mc:")) > 0) {
        switch (option) {
	    case '?':
	        printf("\n\n**** Nieznana opcja: %c ****\n\n", optopt);
                Parametry();
                exit(0);
                break;

	    case 'k':    /* nie testuj kabla poczeniowego */
                Mode_Cable = FALSE;
                break;

            case 'w':    /* wycz UPS  */
                Mode_Kill = TRUE;
                break;

            case 't':    /* tryb testowy  */
                Mode_Test = TRUE;
                break;

            case 'l':    /* wyczaj komputer tylko wtedy, gdy LBAT  */
                Mode_LowBat = TRUE;
                break;

            case 'i':   /* prbkowanie UPSa, < 0 oznacza pojedyczy pomiar */
                if (optarg) {
                    Mode_Delay = atoi (optarg);
                    if (Mode_Delay < 0) {
                        Mode_Test = TRUE;
                        Mode_NoMaster = TRUE;
                    }
                }
                else
                    LogError ("main", "Zly okres probkowania stanu UPSa");
                break;

            case 's':    /* tryb slave  */
                Mode_Slave = TRUE;
                break;

            case 'h':
                Parametry();
                exit(0);
                break;

            case 'p':    /* port IP Port dla Mastera (serwera)  */
                if (optarg)
                    Mode_Port = atoi (optarg);
                else
                    LogError ("main", "Zly numer portu");
                break;

            case 'm':    /* wyczenie trybu Master */
                Mode_NoMaster = TRUE;
                break;

            case 'c':
                if (optarg) {
                    if ((Mode_Count = atoi (optarg)) < 0) {
                        Mode_Count = 0;
                        LogError ("main", "Licznik musi byc nieujemny");
                    }
                }
                else
                    LogError ("main", "Zla wartosc licznika");
                break;
            }
        }

    unlink (PWRSTAT);

    if (optind >= argc) {
        LogError ("main", "Nie podano portu komunikacyjnego");
        Parametry();
        exit(1);
    }

    /*  
     *  Inicjacja portu szergowego i gniazdek Unixowych
     */

    if (Mode_Slave) {
	if ((inaddr = NET_GetServerAddr (argv [optind])) == (unsigned) -1)
            exit (1);
    }
    else {
      if ((fd = UPS_Open (argv [optind])) < 0)
	exit (1);

      /* inicjacja portu przez podanie +12V na RTS lub wyczenie UPSa */
      if (Mode_Kill)
          ups_setup (fd, 1);
      else
          ups_setup (fd, 0);
    }

    /*  
     * A teraz zostaj demonem jeeli nie jest to tryb testowy
     */

    if (!Mode_Test)
      switch (fork ())
      {
        case 0: /* jestem potomkiem */
	    setsid ();   /* tworz now sesj i zostaj jej przywdc */
            chdir ("/");
            umask (0);
            ++IsDaemon;
            break;
            
        case -1: /* bd funkcji fork() */
            LogError ("daemon", "Blad funkcji fork()");
            exit (1);
            
        default: /* Jestem przodkiem */
            exit (0);
      }


    /*  
     * Pocztek monitorowania UPSa
     */

    while (1) {
        if ( Mode_Delay < 0 ) {

            /*  
	     * Pojedyczy test UPSa, zwraca status UPSa jako kod > 100
             */

            if (Mode_Slave)
                exit (NET_Check(inaddr, Mode_Port) + 100);
            else
                exit (ups_check(fd) + 100);
        }
        else {

	      /*  
	       * Tryb normalnej pracy
	       */

	      if ( !Mode_Slave )
		UPS_Status(ups_check (fd));                
	      else
                UPS_Status(NET_Check (inaddr, Mode_Port));
             
	      if ( Mode_NoMaster || Mode_Slave )
                sleep(Mode_Delay);
	      else
                NET_Serve(Mode_Port, Mode_Delay);
        }
   } /* while (1) */

       /*  
        * Powany bd, nigdy nie powinien wystpi
        */

        LogError("main", "Powazny blad aplikacji, przerwanie programu");
        return 1;

} /* main() */


